/* Wingate Monitor Bot -- (c) 2001 Darcy Grexton
   Based on wgmon, (c) 1999-2001 Lucas Madar & DALnet
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "h.h"
#include "settings.h"
#include "struct.h"

#define HASHSIZE 65269		/* Largest prime < 65536-256 */
#define PRIMENUM 607		/* Some random prime, change for optimization */

/* hvalues lookup table:
 * f(x) = (x * primenum) % HASHSIZE
 * 5-10% faster than doing this calculation every time.. 
 */
static unsigned short hvalues[HASHSIZE + 256];
static struct client *clhashtable[HASHSIZE];

Client *headclient;
long cstructs;

void setup_hash()
{
    unsigned short i;

    bzero ((char *)&clhashtable, sizeof(int) * HASHSIZE);

    for (i = 0; i < HASHSIZE + 256; i++)
	hvalues[i] = (i * PRIMENUM) % HASHSIZE;

    headclient = NULL;
    cstructs = 0;
}

static unsigned short hash_name (register char *n)
{
    register unsigned short hv = tolower(*n++);

    while (*n)
	hv = hvalues[hv] + tolower(*n++);

    return hvalues[hv];
}

void del_host_hash (char *host)
{
    unsigned short hashv = hash_name(host);
    Client *ncl = clhashtable[hashv];
    Client *cp;

    if (!ncl)		/* client not in bucket.. */
	return;

    for (cp = ncl; cp; cp = cp->hnext)
    {
	if (!(strcasecmp (host, cp->host)))
	    break;
    }

    if (!cp)		/* other clients in bucket, but not "host".. */
	return;

    cp->cnt--;
    cp->last1 = 0;		/* a client has quit: reset clone detection for this host. */

    if (cp->cnt < 0)	/* can't have a negative amount of clients, can we? :) */
	cp->cnt = 0;

    if (cp->cnt)		/* have more clients on this host, keep it */
	return;

    cp->flags &= KEEP_FLAGS;	/* remove all flags that we don't want to keep
				   when the client is gone. */

    if (cp->flags)
	return;		/* This host still has flags. */

    if (cp->trigger != DEFAULT_TRIGGER_LEVEL)
	return;		/* this host is triggered to something else. */

    if (cp->hprev)		/* not at top of bucket */
	cp->hprev->hnext = cp->hnext;
    else if ((clhashtable[hashv] = cp->hnext))
	clhashtable[hashv]->hprev = NULL;

    if (cp->hnext)
	cp->hnext->hprev = cp->hprev;

    if (cp->prev)		/* in the middle of the linked list .. */
	cp->prev->next = cp->next;
    else if ((headclient = cp->next))
	headclient->prev = NULL;

    if (cp->next)
	cp->next->prev = cp->prev;

    cstructs--;

    free (cp);
}

int find_host_hash_cnt (char *host)
{
    unsigned short hashv = hash_name(host);
    Client *cp = clhashtable[hashv];

    if (cp)
    {
	for (; cp; cp = cp->hnext)
	{
	    if (!(strcasecmp(host, cp->host)))
	    return cp->cnt;
	}
    }

    return 0;
}

Client *find_host_hash (char *host)
{
    unsigned short hashv = hash_name(host);
    Client *cp = clhashtable[hashv];

    if (cp)
    {
	for (; cp; cp = cp->hnext)
	{
	    if (!(strcasecmp(host, cp->host)))
	    return cp;
	}
    }

    return NULL;
}

Client *add_host_hash (char *host)
{
    unsigned short hashv = hash_name(host);
    Client *ncl = clhashtable[hashv];
    Client *cp;

    actualuser++;

    if (ncl)
    {
	for (cp = ncl; cp; cp = cp->hnext)
	{
	    if (!(strcasecmp(host, cp->host)))
	    {
		cp->cnt++;
		numclones++;
		return cp;
	    }
	}
    }

    cp = (Client *) malloc(sizeof(Client));
    memset(cp, '\0', sizeof(Client));
    strcpy(cp->host, host);
    cp->cnt = 1;
    cp->trigger = DEFAULT_TRIGGER_LEVEL;

    if (ncl)
	ncl->hprev = cp;

    cp->hnext = ncl;
    cp->hprev = NULL;
    clhashtable[hashv] = cp;

    cp->prev = NULL;
    cp->next = headclient;
    headclient = cp;

    if (cp->next)
	cp->next->prev = cp;

    cstructs++;

    return cp;
}

void match_host (char *to, char *mask)
{
    Client *cp = headclient;
    int cnt = 0;
    int x;

    while (cp)
    {
	if ((x = match(mask, cp->host)) == 0)
	    cnt++;

	cp = cp->next;
    }

    ssend ("PRIVMSG %s :%d users match %s\n", to, cnt, mask);
}

void hash_stat(char *to)
{
    int i, j;
    int usedb = 0;
    int deepb = 0;
    int twod = 0;
    int threed = 0;
    int fourd = 0;
    Client *ncl;

    ssend("PRIVMSG %s :Making hash stats..\n", to);

    for (i = 0; i < HASHSIZE; i++)
    {
	if (!(ncl = clhashtable[i]))
	    continue;

	usedb++;
	j = 1;

	while ((ncl = ncl->hnext))
	{
	    j++;

	    switch (j)
	    {
		case 2:
		    twod++;
		    break;
		case 3:
		    threed++;
		    break;
		case 4:
		    fourd++;
		    break;
		default:
		    break;
	    }
	}

	if (j > deepb)
	    deepb = j;
    }

    ssend("PRIVMSG %s :Buckets: %d. Used: %d\n", to, HASHSIZE, usedb);
    ssend("PRIVMSG %s :Deepest: %d. Two: %d. Three: %d. Four: %d\n", to,
	deepb, twod, threed, fourd);
}

void save_triggers ()
{
    Client *ncl = headclient;
    FILE *fp;

    fp = fopen(TRIGGERFILE, "w");

    if (!fp)
	return;

    while (ncl)
    {
	if (ncl->trigger != DEFAULT_TRIGGER_LEVEL)
	    fprintf(fp, "%s %lu\n", ncl->host, ncl->trigger);

	ncl = ncl->next;
    }

    fclose(fp);
}

void munge_rn (char *x)
{
    char *y;

    y = strchr(x, '\n');

    if (y)
	*y = '\0';

    y = strchr(x, '\r');

    if (y)
	*y = '\0';
}

void load_triggers ()
{
    FILE *fp;
    char buf[256];
    Client *cp;
    char *u, *val;
    int rval;

    fp = fopen (TRIGGERFILE, "r");
    if (!fp)
	return;

    while (fgets(buf, 256, fp))
    {
	munge_rn(buf);
	u = buf;
	val = index(buf, ' ');

	if (!val)
	    continue;

	*val = '\0';
	val++;

	if (*u == '#' || *u == '\0')
	    continue;

	rval = strtoul(val, NULL, 0);
	if (!rval || rval == DEFAULT_TRIGGER_LEVEL)
	    continue;

	cp = add_host_hash(u);
	cp->cnt--;
	cp->trigger = rval;
    }

    fclose(fp);
}

void load_ignore ()
{
    FILE *fp;
    char buf[256];
    Client *cp;
    char *u, *val;

    fp = fopen(IGNOREFILE, "r");
    if (!fp)
	return;

    while (fgets(buf, 256, fp))
    {
	munge_rn(buf);
	u = buf;
	val = index(buf, ' ');

	if (val)
	    *val = '\0';

	if (*u == '#' || *u == '\0')
	    continue;

	cp = add_host_hash(u);
	cp->cnt--;	/* kludge */
	cp->flags |= HOST_IGNORE;
    }

    fclose(fp);
}

void load_autoban ()
{
    FILE *fp;
    char buf[256];
    Client *cp;
    char *u, *val;

    fp = fopen(AUTOBANFILE, "r");

    if (!fp)
	return;

    while (fgets(buf, 256, fp))
    {
	munge_rn(buf);
	u = buf;
	val = index(buf, ' ');

	if (val)
	    *val = '\0';

	if (*u == '#' || *u == '\0')
	    continue;

	cp = add_host_hash(u);
	cp->cnt--;	/* kludge */
	cp->flags |= HOST_AUTOB;
    }

    fclose(fp);
}

void kill_hash ()
{
    Client *ncl = headclient;
    Client *nncl;

    while (ncl)
    {
	nncl = ncl->next;

	free(ncl);

	ncl = nncl;
    }

    headclient = NULL;
    cstructs = 0;
    bzero((char *)&clhashtable, sizeof(int) * HASHSIZE);
}
