/* Wingate Monitor Bot -- (c) 2001 Darcy Grexton
   Based on wgmon, (c) 1999-2001 Lucas Madar & DALnet

   This program is free. GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <stdarg.h>
#include "h.h"
#include "struct.h"
#include "msg.h"
#include "settings.h"

/* LOCAL VARIABLES! */
static unsigned char socksid[12];	/* SOCKS buffer to send to clients */
static unsigned char socks5id[12];	/* SOCKSv5 buffer to send to clients */
static char sbbuffer[512];	/* My IRC text buffer */
static int sblen;		/* Length of said buffer */
static int s_sock;		/* Socket connected to IRC server */
FILE *logf;			/* File I log stuff to, etc. */
static char LOGFILE[MAX_PATH];	/* Name of said file.. */
char CONFFILE[MAX_PATH];	/* My configuration file name */
char TRIGGERFILE[MAX_PATH];	/* My trigger file name */
char AUTOBANFILE[MAX_PATH];	/* My "autoban" file */
char IGNOREFILE[MAX_PATH];	/* My host ignore file */

static struct fdi locals[MAX_OPEN];	/* better than dynamically allocating them? */
static int highfd;		/* Highest fd in use */
static int numopen;		/* Number of open sockets */
static time_t startime;		/* The time I started at (for uptime) */
static time_t lasttalk;		/* how long has it been since I last got something from the server? */

static int gotsignal = 0;	/* got a signal */
static int gotsighup = 0;	/* got a kill -HUP */
static int gotsigterm = 0;	/* got a kill -TERM */

int pingstate = 0;		/* I sent a PING, waiting for some reply? */

unsigned long acons4 = 0;	/* Attempted connections */
unsigned long acons5 = 0;	/* Attempted connections */
unsigned long aconwg = 0;	/* Attempted connections */
unsigned long cons4 = 0;	/* Successful connections s4 */
unsigned long cons5 = 0;	/* Successful connections s5 */
unsigned long conwg = 0;	/* Successful connections wg */
unsigned long opns4 = 0;	/* Open socks4 */
unsigned long opns5 = 0;	/* Open socks5 */
unsigned long opnwg = 0;	/* Open wingate */
unsigned long numclones = 0;	/* Number of clones */
unsigned long actualuser = 0;	/* Number of actual users (not clones) */

unsigned long currenttime;	/* cache time() */

static char *fmt_time(unsigned long t)
{
	static char tstr[64];
	unsigned long h, m, s, d, w, r;

	h = m = s = d = w = r = 0;

	if (t > 604800) {
		w = t / 604800;
		r = t % 604800;
	} else
		r = t;

	if (r > 86400) {
		d = r / 86400;
		r %= 86400;
	}

	if (r > 3600) {
		h = r / 3600;
		r %= 3600;
	}

	if (r > 60) {
		m = r / 60;
		r %= 60;
	}

	s = r;


	sprintf(tstr, "%lu wks %lu days %lu hrs %lu min %lu sec", w, d, h, m,
		s);

	return tstr;
}

int ssend(char *pattern, ...)
{
	va_list vl;
	int slen, x;
	static char sendbuf[2048];

	va_start(vl, pattern);
	slen = vsprintf(sendbuf, pattern, vl);
	x = send(s_sock, sendbuf, slen, 0);
	va_end(vl);

#ifdef CDEBUG
	printf("s: %s", sendbuf);
#endif

	return x;
}

int getsockerr(int fd)
{
	int err = 0, len = sizeof(err);

	if (getsockopt(fd, SOL_SOCKET, SO_ERROR, (char *)&err, &len))
		return EINVAL;

	return err;
}


void maxopenfull()
{
	ssend
	    ("LOCOPS :WARNING: Maximum number of open connections reached. Increase MAX_OPEN.\n");
}

void parse(char *buf, struct Message *mptr)
{
	int i;
	char *s, *tmp;

#define PARAS 20

	static char *para[PARAS + 1];	/* max 20 parms, 1 null */

#ifdef PARSEDEBUG
	printf("P: %s\n", buf);
#endif

	para[0] = NULL;

	if (*buf == ':') {
		s = &buf[1];
		if (!(*s))
			return;

		para[0] = s;

		s = strchr((char *)&buf[1], ' ');
		if (!s)		/* really shouldn't happen.. */
			return;

		*s++ = '\0';

	} else
		s = &buf[0];


	tmp = strchr(s, ' ');
	if (!tmp)		/* another no-no.. */
		return;

	*tmp++ = '\0';

	for (; mptr->cmd; mptr++)
		if (strcmp(mptr->cmd, s) == 0)
			break;

	if (!mptr->cmd)		/* dunno the command, exit quietly.. */
		return;

	s = tmp;
	i = 0;

	while (1) {
		while (*s == ' ')
			*s++ = '\0';

		if (*s == '\0')
			break;

		if (*s == ':') {
			para[++i] = s + 1;
			break;
		}

		para[++i] = s;

		if (i >= PARAS)
			break;

		for (; *s != ' ' && *s; s++) {	/* nada */
		}
	}

	(*mptr->func) (i, para);

}

void handlemainpacket(char *buffer, int len)
{
	char *buf1, *buf2;

	buf1 = sbbuffer + sblen;
	buf2 = buffer;

	while (--len >= 0) {
		register char c = (*buf1 = *buf2++);

		if (c < '\16' && (c == '\n' || c == '\r')) {
			if (buf1 == sbbuffer)
				continue;

			*buf1 = '\0';

			sblen = 0;

			parse(sbbuffer, msgtab);

			buf1 = sbbuffer;
		} else if (buf1 < sbbuffer + (sizeof(sbbuffer) - 1))
			buf1++;
	}

	sblen = buf1 - sbbuffer;
}

char *get_stat(int s)
{
	switch (locals[s].status) {
	case S_S4_NEGOTIATE:
	case S_S4_REPLYWAIT:
		return "socks4";

	case S_S5_NEGOTIATE:
	case S_S5_REPLYWAIT:
		return "socks5";

	case S_WG_NEGOTIATE:
	case S_WG_REPLYWAIT:
	case S_WG_INFOWAIT:
		return "wingate";

	default:
		return "UNKNOWN!";
		break;
	}
}

#ifdef SOCKDEBUG
char *get_debugstat(int s)
{
	switch (locals[s].status) {
	case S_S4_NEGOTIATE:
		return "socks4_negotiate";
	case S_S4_REPLYWAIT:
		return "socks4_wait";

	case S_S5_NEGOTIATE:
		return "socks5_negotiate";
	case S_S5_REPLYWAIT:
		return "socks5_wait";

	case S_WG_NEGOTIATE:
		return "wingate_negotiate";
	case S_WG_REPLYWAIT:
		return "wingate_wait";
	case S_WG_INFOWAIT:
		return "wingate_info";

	case S_FREE:
		return "FREE";

	default:
		return "UNKNOWN!";
	}
}

#endif

void loguser(int i)
{
	struct in_addr ia;
	char *nick = locals[i].n;
	char *user = locals[i].u;
	char *host = locals[i].h;
	char *src = get_stat(i);
	char *ip;
	char tstr[128];
	time_t now = currenttime;

	ia.s_addr = locals[i].ip;
	ip = inet_ntoa(ia);

	if (!(locals[i].flags & F_SHOW))
		ssend("PRIVMSG %s :%s: %s!%s@%s\n", MYCHAN, src, nick, user,
		      host);
	else {
		ssend("PRIVMSG %s :Open %s: %s [%s]\n",	/* with this, target is "nick" .. it */
		      nick, src, host, ip);	/* had been requested by someone */

		return;		/* don't log requested checks, either. */
	}


	strftime(tstr, 128, "%Y%m%d%H%M%S", gmtime(&now));

	fprintf(logf, "%s:%s:%s:%s:%s:%s\n", tstr, src, nick, user, host, ip);
	fflush(logf);
}

void killuser(int i)
{
	char *host = locals[i].h;

	/* Killing the user gives out the proxy IP address, and that's
	 * generally a bad idea (tm) */

	ssend("KLINE %d *@%s :%s\n", KLINETIME, host, KREASON);
}

int connect_irc()
{
	int s;
	struct sockaddr_in sin;

	while (1) {
		s = socket(AF_INET, SOCK_STREAM, 0);
		if (s < 0) {
			sleep(60);
			continue;
		}

		memset(&sin, 0, sizeof(struct sockaddr_in));

		if (BINDIP) {
			sin.sin_family = AF_INET;
			sin.sin_port = 0;
			sin.sin_addr.s_addr = BINDIP;

			bind(s, (struct sockaddr *)&sin,

			     sizeof(struct sockaddr_in));
		}

		sin.sin_family = AF_INET;
		sin.sin_port = htons(SERVPORT);
		sin.sin_addr.s_addr = inet_addr(SERVIP);

		if (connect
		    (s, (struct sockaddr *)&sin,
		     sizeof(struct sockaddr_in))) {
			close(s);
			sleep(10);
		} else
			break;
	}

	s_sock = s;

	ssend("NICK %s\nUSER %s\n", NICK, USER);

/*	fcntl(s_sock, F_SETFL, O_NONBLOCK);*/

/* if this socket blocks, we're useless anyway.. */

/* otherwise, I need buffering routines. uck. */

	return 0;
}

void do_s4(char *k, char *u, unsigned long ip, char *n, int show)
{
	struct sockaddr_in sin;
	int s;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return;

	if (s >= MAX_OPEN) {
		maxopenfull();
		close(s);
		return;
	}

	memset(&sin, 0, sizeof(struct sockaddr_in));

	if (BINDIP) {
		sin.sin_family = AF_INET;
		sin.sin_port = 0;
		sin.sin_addr.s_addr = BINDIP;

		if (bind
		    (s, (struct sockaddr *)&sin,
		     sizeof(struct sockaddr_in))) {
			ssend
			    ("LOCOPS :\002BIND() FAILURE\002! Reverting to inaddr_any!\n");
			BINDIP = 0;
		}
	}

	sin.sin_family = AF_INET;
	sin.sin_port = htons(1080);
	sin.sin_addr.s_addr = ip;

	fcntl(s, F_SETFL, O_NONBLOCK);

	if (connect(s, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))
	    && errno != EINPROGRESS) {
		/* This is pretty much an error, ie, connect should ALWAYS return EINPROGRESS! */
		/* apparently it can fail for a connect to localhost. */
		/* ssend("PRIVMSG %s :Failed connect to %s (%s)\n", MYCHAN, n, strerror(errno)); */
		close(s);
		return;
	}

	if (highfd < s)
		highfd = s;

	numopen++;
	acons4++;

	switch (show) {
	case C_SHOW:
		locals[s].flags = F_SHOW | F_MANUAL;
		/* don't show, don't kill the user, don't progress to wingate check */
		break;
	case C_NORM:
	default:
		locals[s].flags = 0;
		break;
	}

	locals[s].status = S_S4_NEGOTIATE;
	locals[s].ip = ip;
	locals[s].ott = currenttime + TIMEOUT;
	strcpy(locals[s].h, n);
	strcpy(locals[s].n, k);
	strcpy(locals[s].u, u);
}

void do_s5(char *k, char *u, unsigned long ip, char *n, int show)
{
	struct sockaddr_in sin;
	int s;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return;

	if (s >= MAX_OPEN) {
		maxopenfull();
		close(s);
		return;
	}

	memset(&sin, 0, sizeof(struct sockaddr_in));

	if (BINDIP) {
		sin.sin_family = AF_INET;
		sin.sin_port = 0;
		sin.sin_addr.s_addr = BINDIP;

		if (bind
		    (s, (struct sockaddr *)&sin,
		     sizeof(struct sockaddr_in))) {
			ssend
			    ("LOCOPS :\002BIND() FAILURE\002! Reverting to inaddr_any!\n");
			BINDIP = 0;
		}
	}

	sin.sin_family = AF_INET;
	sin.sin_port = htons(1080);
	sin.sin_addr.s_addr = ip;

	fcntl(s, F_SETFL, O_NONBLOCK);

	if (connect(s, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))
	    && errno != EINPROGRESS) {
		/* This is pretty much an error, ie, connect should ALWAYS return EINPROGRESS! */
		/* apparently it can fail for a connect to localhost. */
		/* ssend("PRIVMSG %s :Failed connect to %s (%s)\n", MYCHAN, n, strerror(errno)); */
		close(s);
		return;
	}

	if (highfd < s)
		highfd = s;

	numopen++;
	acons5++;

	switch (show) {
	case C_SHOW:
		locals[s].flags = F_SHOW | F_MANUAL;
		/* don't show, don't kill the user, don't progress to wingate check */
		break;
	case C_NORM:
	default:
		locals[s].flags = 0;
		break;
	}

	locals[s].status = S_S5_NEGOTIATE;
	locals[s].ip = ip;
	locals[s].ott = currenttime + TIMEOUT;
	strcpy(locals[s].h, n);
	strcpy(locals[s].n, k);
	strcpy(locals[s].u, u);
}

void do_wg(char *k, char *u, unsigned long ip, char *n, int show)
{
	struct sockaddr_in sin;
	int s;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return;

	if (s >= MAX_OPEN) {
		maxopenfull();
		close(s);
		return;
	}

	memset(&sin, 0, sizeof(struct sockaddr_in));

	if (BINDIP) {
		sin.sin_family = AF_INET;
		sin.sin_port = 0;
		sin.sin_addr.s_addr = BINDIP;

		if (bind
		    (s, (struct sockaddr *)&sin,
		     sizeof(struct sockaddr_in))) {
			ssend
			    ("LOCOPS :\002BIND() FAILURE\002! Reverting to inaddr_any!\n");
			BINDIP = 0;
		}
	}

	sin.sin_family = AF_INET;
	sin.sin_port = htons(23);
	sin.sin_addr.s_addr = ip;

	fcntl(s, F_SETFL, O_NONBLOCK);

	if (connect(s, (struct sockaddr *)&sin, sizeof(struct sockaddr_in))
	    && errno != EINPROGRESS) {
		/* This is pretty much an error, ie, connect should ALWAYS return EINPROGRESS! */
		/* ssend("PRIVMSG %s :Failed connect to %s (%s)\n", MYCHAN, n, strerror(errno)); */
		close(s);
		return;
	}

	if (highfd < s)
		highfd = s;

	numopen++;
	aconwg++;

	switch (show) {
	case C_SHOW:
		locals[s].flags = F_MANUAL | F_SHOW;
		/* don't progress to anything... */
		break;
	case C_NORM:
	default:
		locals[s].flags = 0;
		break;
	}

	locals[s].status = S_WG_NEGOTIATE;
	locals[s].ip = ip;
	locals[s].ott = currenttime + TIMEOUT;
	strcpy(locals[s].h, n);
	strcpy(locals[s].n, k);
	strcpy(locals[s].u, u);
}

void show_memstats(char *target)
{
	long hashsize = cstructs * sizeof(struct client);
	long memsize = MAX_OPEN * sizeof(struct fdi);

	ssend("PRIVMSG %s :%d host hash structs, %d bytes.\n", target,
	      cstructs, hashsize);
	ssend("PRIVMSG %s :%d wg/s4 structs, %d bytes.\n", target, MAX_OPEN,
	      memsize);
}

void show_stats(char *target)
{
	time_t now = currenttime;

	ssend("PRIVMSG %s :I have %d open sockets. Uptime: %s\n", target,
	      numopen, fmt_time(now - startime));
	ssend("PRIVMSG %s :Found %d socks5 servers. %d open in %d users.\n",
	      target, cons5, opns5, acons5);
	ssend("PRIVMSG %s :Found %d socks4 servers, %d open in %d users.\n",
	      target, cons4, opns4, acons4);
	ssend("PRIVMSG %s :Found %d wingates, %d open in %d users.\n", target,
	      conwg, opnwg, aconwg);
	ssend("PRIVMSG %s :Number of connects: %d. Number of clones: %d\n",
	      target, actualuser, numclones);

#ifdef SOCKDEBUG
	{
		FILE *fp;
		int i;
		time_t c = time(NULL);


		fp = fopen("debug.out", "w");
		if (!fp)
			return;

		fprintf(fp, "I have %d open sockets. Uptime: %s\n", numopen,
			fmt_time(now - startime));
		fprintf(fp,
			"Found %lu socks5 servers. %lu open in %lu users.\n",
			cons5, opns5, acons5);
		fprintf(fp,
			"Found %lu socks4 servers, %lu open in %lu users.\n",
			cons4, opns4, acons4);
		fprintf(fp, "Found %lu wingates, %lu open in %lu users.\n",
			conwg, opnwg, aconwg);
		fprintf(fp,
			"Number of connects: %lu. Number of clones: %lu\n",
			actualuser, numclones);

		for (i = 0; i < MAX_OPEN; i++) {
			fprintf(fp, "[%lu] S: %s F: %d H: %s OTT: %lu\n", i,
				get_debugstat(i), locals[i].flags,
				locals[i].h, c - locals[i].ott);
		}

		fclose(fp);
	}
#endif
}

void build_socksid()
{
	unsigned long theip, netip;

	netip = inet_addr(SCIP);
	theip = htonl(netip);

	socksid[0] = 4;		// socks 4
	socksid[1] = 1;		// connect request
	socksid[2] = (((unsigned short)SCPORT) >> 8) & 0xFF;	// hiword of port
	socksid[3] = (((unsigned short)SCPORT) & 0xFF);	// loword of port
	socksid[4] = (theip >> 24) & 0xFF;
	socksid[5] = (theip >> 16) & 0xFF;
	socksid[6] = (theip >> 8) & 0xFF;
	socksid[7] = (theip & 0xFF);
	socksid[8] = 0;

	socks5id[0] = 5;	/* socks v5 */
	socks5id[1] = 1;	/* 1 auth method */
	socks5id[2] = 0;	/* no authentication */
}

int check_dead()
{
	static time_t last_check;

	if (pingstate == 1) {
		if (last_check + 120 < currenttime)
			return 0;
		/* ack, the server is dead, it seems. */
		return -1;
	}

	last_check = currenttime;
	pingstate = 0;
	ssend("PING :%s\n", NICK);

	return 0;
}

int irc_loop()
{
	fd_set rfd;
	fd_set wfd;
	int nfds, i;
	struct timeval tv;
	time_t now;
	int retval;

	lasttalk = time(NULL);	/* we just connected, this is the server "talking" */

	pingstate = 0;

	numopen = 0;
	sblen = 0;

	build_socksid();

	bzero((char *)&locals[0], sizeof(struct fdi) * MAX_OPEN);

	highfd = 0;

	while (1) {
		FD_ZERO(&rfd);
		FD_ZERO(&wfd);
		FD_SET(s_sock, &rfd);

		for (i = 0; i <= highfd; i++) {
			switch (locals[i].status) {
			case S_S4_NEGOTIATE:
			case S_S5_NEGOTIATE:
			case S_WG_NEGOTIATE:
				FD_SET(i, &wfd);
			case S_S4_REPLYWAIT:
			case S_S5_REPLYWAIT:
			case S_WG_REPLYWAIT:
			case S_WG_INFOWAIT:
				FD_SET(i, &rfd);
			default:
				break;
			}
		}

		tv.tv_sec = 10;
		tv.tv_usec = 0;

		nfds = select(FD_SETSIZE, &rfd, &wfd, NULL, &tv);

		/* SIGNAL STUFF */

		if (gotsignal) {
			gotsignal = 0;

			if (gotsighup) {
				int lc;

				ssend
				    ("PRIVMSG %s :got SIGHUP, reloading...\n",
				     MYCHAN);
				fclose(logf);
				logf = fopen(LOGFILE, "a");

				lc = loadconf(1);

				if (lc == CONF_FAILURE) {
					ssend
					    ("PRIVMSG %s :Error loading conf!\n",
					     MYCHAN);
				} else {

					if (lc & CONF_NICKCHG)
						ssend("NICK %s\n", NICK);

					if (lc & CONF_CHANCHG)
						ssend("JOIN 0\nJOIN %s\n",
						      MYCHAN);

					if (lc & CONF_UPDATES4)
						build_socksid();

				}

				gotsighup = 0;
			} else if (gotsigterm) {
				ssend("PRIVMSG %s :Got TERM signal, dying.\n",
				      MYCHAN);
				ssend("QUIT :TERM, dying off.\n");
				retval = 0;
				sleep(2);
				break;
			}
		}

		/* END SIGNAL STUFF */

		if (nfds < 0)	/* probably a signal.. */
			continue;

		currenttime = now = time(NULL);

		if (FD_ISSET(s_sock, &rfd)) {
			int len;
			static char getbuf[2048];

			if ((len = recv(s_sock, getbuf, sizeof(getbuf), 0)) <=
			    0) {
				close(s_sock);
				retval = 1;
				break;
			}

			lasttalk = now;

			handlemainpacket(getbuf, len);
		}

		if (lasttalk + 60 < now) {
			if (check_dead()) {
				retval = 0;
				break;
			}
			lasttalk = now;
		}

		for (i = 0; i <= highfd; i++) {
			if (locals[i].status) {
//////////////////////////////////////////////////////////////
/////// WRITE LOOP                               ////////////
/////////////////////////////////////////////////////////////

				if (FD_ISSET(i, &wfd)) {
					if (locals[i].status ==
					    S_S5_NEGOTIATE) {
						cons5++;

						if (send(i, socks5id, 3, 0) !=
						    3) {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server. (Write negotiate: %s)\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i),
								     errno ?
								     strerror
								     (errno) :
								     "Dropped");

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;

							if (!
							    (locals[i].
							     flags &
							     F_MANUAL))
							    do_wg(locals[i].n,
								  locals[i].u,
								  locals[i].
								  ip,
								  locals[i].h,
								  C_NORM);

							continue;
						}
						locals[i].ott = now + TIMEOUT;
						locals[i].status =
						    S_S5_REPLYWAIT;
						locals[i].flags |= F_OPEN1080;	/* port 1080 is open! */
					} else if (locals[i].status ==
						   S_S4_NEGOTIATE) {
						cons4++;

						if (send(i, socksid, 9, 0) !=
						    9) {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server. (Write negotiate: %s)\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i),
								     errno ?
								     strerror
								     (errno) :
								     "Dropped");

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;


							if (!
							    (locals[i].
							     flags &
							     F_MANUAL))
							    do_wg(locals[i].n,
								  locals[i].u,
								  locals[i].
								  ip,
								  locals[i].h,
								  C_NORM);

							continue;
						}
						locals[i].ott = now + TIMEOUT;
						locals[i].status =
						    S_S4_REPLYWAIT;
						locals[i].flags |= F_OPEN1080;	/* port 1080 is open! */
					} else if (locals[i].status ==
						   S_WG_NEGOTIATE) {
						conwg++;

						if (send(i, "\r\n", 2, 0) !=
						    2) {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server. (Write negotiate: %s)\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i),
								     errno ?
								     strerror
								     (errno) :
								     "Dropped");

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;
						}
						locals[i].ott = now + TIMEOUT;
						locals[i].status =
						    S_WG_REPLYWAIT;
					}
				}
//////////////////////////////////////////////////////////////
/////// READ LOOP                                ////////////
/////////////////////////////////////////////////////////////

				if (FD_ISSET(i, &rfd)) {
					int len;
					unsigned char sbuf[64];

					len = recv(i, sbuf, 64, 0);

					/* Connection gave a read error after connect() -- call must have failed. */
					if (len <= 0 && IsNegotiating(i)) {
						struct in_addr ia;
						char stat = locals[i].status;

						ia.s_addr = locals[i].ip;

						if (locals[i].flags & F_SHOW) {
							int err =

							    getsockerr(i);
							ssend
							    ("PRIVMSG %s :%s is not an open %s server. (Negotiate: %s)\n",
							     locals[i].n,
							     locals[i].h,
							     get_stat(i),
							     err ?
							     strerror(err) :
							     "Dropped");
						}
						close(i);
						numopen--;
						locals[i].status = S_FREE;

						/* Hairy:
						 * If we errored in s(4/5) connect(), do a wingate check */
						if (
						    (stat == S_S4_NEGOTIATE
						     || stat ==
						     S_S5_NEGOTIATE)
						    && !(locals[i].
							 flags & F_MANUAL))
							    do_wg(locals[i].n,
								  locals[i].u,
								  locals[i].
								  ip,
								  locals[i].h,
								  C_NORM);

						continue;
					}
					/* Read error while waiting.. drop it. */
					if (len <= 0 && IsWaiting(i)) {
						struct in_addr ia;
						unsigned long flags =
						    locals[i].flags;
						char stat = locals[i].status;

						ia.s_addr = locals[i].ip;

						if (locals[i].flags & F_SHOW) {
							int err =

							    getsockerr(i);
							ssend
							    ("PRIVMSG %s :%s is not an open %s server. (Wait: %s)\n",
							     locals[i].n,
							     locals[i].h,
							     get_stat(i),
							     err ?
							     strerror(err) :
							     "Remote host dropped connection");
						}
						close(i);
						numopen--;
						locals[i].status = S_FREE;

						if (stat == S_S5_REPLYWAIT
						    && (flags & F_OPEN1080)) {
							do_s4(locals[i].n,
							      locals[i].u,
							      locals[i].ip,
							      locals[i].h,
							      (flags &
							       F_MANUAL) ?
							      C_SHOW :
							      C_NORM);
						} else
						    if (
							(stat ==
							 S_S4_REPLYWAIT
							 || stat ==
							 S_S5_REPLYWAIT)
							&& !(flags &
							     F_MANUAL))
							    do_wg(locals[i].n,
								  locals[i].u,
								  locals[i].
								  ip,
								  locals[i].h,
								  C_NORM);

						continue;
					}

					/* Got info, in wait state -- socks5 reply! */
					if (locals[i].status ==
					    S_S5_REPLYWAIT) {
						if (len >= 2 && sbuf[0] == 5
						    && sbuf[1] == 0) {	/* open s5 */
							loguser(i);
							killuser(i);

							opns5++;

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;
						} else {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server. (socks5 reply %d %d)\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i),
								     sbuf[0],
								     sbuf[1]);
							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;

							if (locals[i].
							    flags & F_OPEN1080
							    && (len < 2
								|| sbuf[1] ==
								91))
								    do_s4
								    (locals
								     [i].n,
								     locals
								     [i].u,
								     locals
								     [i].ip,
								     locals
								     [i].h,
								     (locals
								      [i].
								      flags &
								      F_MANUAL)
								     ? C_SHOW
								     :
								     C_NORM);
							continue;
						}
					} else if (locals[i].status ==
						   S_S4_REPLYWAIT) {
						if (len >= 2 && sbuf[1] == 90) {
							/* OPEN SOCKS4 PROXY! */

							loguser(i);
							killuser(i);

							opns4++;

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;
						} else {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server (%d).\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i),
								     sbuf[1]);
							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;
						}
					} else /* data from wingate.. */
				    if (locals[i].status == S_WG_REPLYWAIT
					|| locals[i].status == S_WG_INFOWAIT) {
						if (strncasecmp
						    (sbuf, "WinGate>", 8) == 0
						    || strncasecmp(sbuf,
								   "Too many connected users - try again later",
								   42) == 0) {

							loguser(i);
							killuser(i);

							opnwg++;

							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;

						} else if (locals[i].status ==
							   S_WG_REPLYWAIT) {
							locals[i].status =
							    S_WG_INFOWAIT;
							continue;
						} else {
							if (locals[i].
							    flags & F_SHOW)
								    ssend
								    ("PRIVMSG %s :%s is not an open %s server.\n",
								     locals
								     [i].n,
								     locals
								     [i].h,
								     get_stat
								     (i));
							close(i);
							numopen--;
							locals[i].status =
							    S_FREE;
							continue;
						}
					}

					ssend
					    ("PRIVMSG %s :[debug] fd %d for reading untouched! stat: %d flags %lu\n",
					     MYCHAN, i, locals[i].status,
					     locals[i].flags);
				}

				/* check for timeouts.. */
				/* Note that if a connect times out, it is NOT WORTH IT to try and move on
				 * to the next 'check' stage, as that will time out 99.9% of the time as well. */
				if (now > locals[i].ott) {
					if (locals[i].flags & F_SHOW)
						ssend
						    ("PRIVMSG %s :%s detect timed out for %s.\n",
						     locals[i].n, get_stat(i),
						     locals[i].h);

					close(i);
					numopen--;
					locals[i].status = S_FREE;
				}
			}
		}

	}

	close(s_sock);

	for (i = 0; i <= highfd; i++)
		if (locals[i].status)
			close(i);

	return retval;
}

void sighup(int x)
{
	gotsignal = 1;
	gotsighup = 1;
}

void sigterm(int x)
{
	gotsignal = 1;
	gotsigterm = 1;
}

void segv(int x)
{
	static struct linger sol;
	static int solsz;

	sbbuffer[sizeof(sbbuffer) - 1] = '\0';

	solsz = sizeof(struct linger);

	sol.l_onoff = 1;
	sol.l_linger = 100000;

	setsockopt(s_sock, SOL_SOCKET, SO_LINGER, (void *)&sol, solsz);

	ssend("PRIVMSG %s :RECIEVED SIGNAL 11!\n", MYCHAN);
	ssend("PRIVMSG %s :sbbuffer: %s\n", MYCHAN, sbbuffer);
	ssend("QUIT :Signal 11, yicky!\n");

	sleep(5);

	close(s_sock);

	abort();
}

int main(int argc, char *argv[])
{
#ifndef CDEBUG
	int xx;
	char PIDFILE[MAX_PATH];
#endif
	char *confname = (argc > 1) ? argv[1] : DEFAULTNAME;

	if (strlen(confname) > (MAX_PATH - 12)) {
		printf("Specified name is too long... eugh.\n");
		return 0;
	}
#ifndef CDEBUG
	sprintf(PIDFILE, "%s.%s", confname, PID_EXT);
#endif

	sprintf(CONFFILE, "%s.%s", confname, CONF_EXT);
	sprintf(LOGFILE, "%s.%s", confname, LOG_EXT);
	sprintf(TRIGGERFILE, "%s.%s", confname, TRIGGER_EXT);
	sprintf(AUTOBANFILE, "%s.%s", confname, AUTOBAN_EXT);
	sprintf(IGNOREFILE, "%s.%s", confname, IGNORE_EXT);

	printf("Wingate Connection Monitor starting up. Version %d.%d.%d\n",
	       MAJOR, MINOR, PATCH);
	printf("By lucas <lucas@negaverse.org>\n");
	printf("Modifications by skold <skold@habber.net>\n\n");

	initconfig();
	if (!loadconf(0))
		exit(0);

#ifndef CDEBUG
	if ((xx = fork())) {	/* fork off and die! */
		FILE *pidf;

		pidf = fopen(PIDFILE, "w");

		if (pidf) {
			fprintf(pidf, "%d\n", xx);
			fclose(pidf);
		}

		printf("Forking off and dying! [pid %d]\n", xx);
		exit(0);
	}
#endif

	logf = fopen(LOGFILE, "a");
	if (!logf) {
		printf("grah! can't open logfile!@#\n");
		exit(0);
	}
#ifndef CDEBUG
	close(0);		/* keel them all! */
	close(1);
	close(2);
#endif

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, sighup);
	signal(SIGTERM, sigterm);
#ifndef CDEBUG
	signal(SIGSEGV, segv);
#endif
	startime = time(NULL);

	setup_hash();

	myserv[0] = '\0';

	while (1) {
		connect_irc();
		if (!irc_loop())
			break;
		fflush(logf);
	}

	fclose(logf);

	unlink(PIDFILE);

	return 0;
}
